from __future__ import annotations
import abc
import re
from abc import ABC
from typing import List, Dict, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique import \
        FieldContentProductionTechnique
    from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
    from clayrs.content_analyzer.exogenous_properties_retrieval import ExogenousPropertiesRetrieval
    from clayrs.content_analyzer.memory_interfaces.memory_interfaces import InformationInterface
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.field_content_production_techniques.field_content_production_technique import \
    OriginalData


class FieldConfig:
    """
    Class that represents the configuration for a single representation of a field. The configuration of a single
    representation is defined by a `FieldContentProductionTechnique` (e.g. an `EmbeddingTechnique`) that will be applied
    to the pre-processed data of said field.

    To specify how to preprocess data, simply specify an `InformationProcessor` in the `preprocessing` parameter.
    Multiple `InformationProcessor` can be wrapped in a list: in this case, the field will be preprocessed by performing
    operations all objects inside the list.
    If preprocessing is not defined, no preprocessing operations will be done on the field data.

    You can use the `id` parameter to assign a custom id for the representation: by doing so the user can freely refer
    to it by using the custom id given, rather than positional integers (which are given automatically by the
    framework).

    There is also a memory_interface attribute which allows to define a data structure where the representation will
    be serialized (e.g. an Index).

    Various configurations are possible depending on how the user wants to represent a particular field:

    * This will produce a field representation using the SkLearnTfIdf technique on the field data
      preprocessed by NLTK by performing stopwords removal, and the name of the produced representation will be
      'field_example'

    ```python
    FieldConfig(SkLearnTfIdf(), NLTK(stopwords_removal=True), id='field_example')
    ```

    * This will produce the same result as above but the id for the field representation defined by this config will
      be set by the ContentAnalyzer once it is being processed (0 integer if it's the first representation specified
      for the field, 1 if it's the second, etc.)

    ```python
    FieldConfig(SkLearnTfIdf(), NLTK())
    ```

    * This will produce a field representation using the SkLearnTfIdf technique on the field data without applying
      any preprocessing operation, but it will not be directly stored in the content, instead it will be
      stored in a index

    ```python
    FieldConfig(SkLearnTfIdf(), memory_interface=SearchIndex(/somedir))
    ```

    * In the following nothing will be done on the field data, it will be represented as is

    ```python
    FieldConfig()
    ```

    Args:
        content_technique: Technique that will be applied to the field in order to produce a complex representation
            of said field
        preprocessing: Single `InformationProcessor` object or a list of `InformationProcessor` objects that will be
            used preprocess field data before applying the `content_technique`
        memory_interface: complex structure where the content representation can be serialized (an Index for
            example)
        id: Custom id that can be used later by the user to easily refer to the representation generated by this config.
            IDs for a single field should be unique! And should only contain '_', '-' and alphanumeric characters
    """

    def __init__(self,
                 content_technique: FieldContentProductionTechnique = OriginalData(),
                 preprocessing: Union[InformationProcessor, List[InformationProcessor]] = None,
                 memory_interface: InformationInterface = None,
                 id: str = None):

        if preprocessing is None:
            preprocessing = []

        if id is not None:
            self._check_custom_id(id)

        self.__content_technique = content_technique
        self.__preprocessing = preprocessing
        self.__memory_interface = memory_interface
        self.__id = id

        if not isinstance(self.__preprocessing, list):
            self.__preprocessing = [self.__preprocessing]

    @property
    def memory_interface(self):
        """
        Getter for the index associated to the field config
        """
        return self.__memory_interface

    @property
    def content_technique(self):
        """
        Getter for the field content production technique of the field
        """
        return self.__content_technique

    @property
    def preprocessing(self):
        """
        Getter for the list of preprocessor of the field config
        """
        return self.__preprocessing

    @property
    def id(self):
        """
        Getter for the id of the field config
        """
        return self.__id

    def _check_custom_id(self, id: str):
        if not re.match("^[A-Za-z0-9_-]+$", id):
            raise ValueError("The custom id {} is not valid!\n"
                             "A custom id can only have numbers, letters and '_' or '-'!".format(id))

    def __str__(self):
        return "FieldConfig"

    def __repr__(self):
        return f'FieldConfig(content_technique={self.__content_technique}, preprocessing={self.__preprocessing}, ' \
               f'memory_interface={self.__memory_interface}, id={self.__id})'


class ExogenousConfig:
    """
    Class that represents the configuration for a single exogenous representation.

    The config allows the user to specify an `ExogenousPropertiesRetrieval` technique to use to expand each content.
    W.r.t `FieldConfig` objects, an `ExogenousConfig` does not refer to a particular field but to the whole content
    itself.

    You can use the `id` parameter to assign a custom id for the representation: by doing so the user can freely refer
    to it by using the custom id given, rather than positional integers (which are given automatically by the
    framework).

    * This will create an exogenous representation for the content by expanding it using DBPedia,
    said representation will be named 'test'
    ```python
    ExogenousConfig(DBPediaMappingTechnique('dbo:Film', 'Title', 'EN'), id='test')
    ```

    * Same as the example above, but since no custom id was assigned, the exogenous representation can be referred to
    only with an integer (0 if it's the first exogenous representation specified for the contents, 1 if it's the second,
    etc.)

    ```python
    ExogenousConfig(DBPediaMappingTechnique('dbo:Film', 'Title', 'EN'))
    ```

    Args:
        exogenous_technique: Technique which will be used to expand each content with data from external sources.
            An example would be the DBPediaMappingTechnique which allows to retrieve properties from DBPedia.
        id: Custom id that can be used later by the user to easily refer to the representation generated by this config.
            IDs for a single field should be unique! And should only contain '_', '-' and alphanumeric characters
    """

    def __init__(self, exogenous_technique: ExogenousPropertiesRetrieval, id: str = None):
        if id is not None:
            self._check_custom_id(id)

        self.__exogenous_technique = exogenous_technique
        self.__id = id

    @property
    def exogenous_technique(self):
        """
        Getter for the exogenous properties retrieval technique
        """
        return self.__exogenous_technique

    @property
    def id(self):
        """
        Getter for the ExogenousConfig id
        """
        return self.__id

    def _check_custom_id(self, id: str):
        if not re.match("^[A-Za-z0-9_-]+$", id):
            raise ValueError("The custom id {} is not valid!\n"
                             "A custom id can only have numbers, letters and '_' or '-'!".format(id))

    def __str__(self):
        return "ExogenousConfig"

    def __repr__(self):
        return f'ExogenousConfig(exogenous_technique={self.__exogenous_technique}, ' \
               f'id={self.__id})'


class ContentAnalyzerConfig(ABC):
    """
    Abstract class that represents the configuration for the content analyzer. The configuration specifies how the
    `Content Analyzer` needs to complexly represent contents, i.e. how to preprocess them and how to represent them

    Args:
        source: Raw data source wrapper which contains original information about contents to process
        id: Field of the raw source which represents each content uniquely.
        output_directory: Where contents complexly represented will be serialized
        field_dict: Dictionary object which contains, for each field of the raw source to process, a FieldConfig object
            (e.g. `{'plot': FieldConfig(SkLearnTfIdf(), 'genres': FieldConfig(WhooshTfIdf()))}`)
        exogenous_representation_list: List of `ExogenousTechnique` objects that will be used to expand each contents
            with data from external sources
        export_json: If set to True, contents complexly represented will be serialized in a human readable JSON, other
            than in a proprietary format of the framework
    """

    def __init__(self, source: RawInformationSource,
                 id: Union[str, List[str]],
                 output_directory: str,
                 field_dict: Dict[str, List[FieldConfig]] = None,
                 exogenous_representation_list: Union[ExogenousConfig, List[ExogenousConfig]] = None,
                 export_json: bool = False):
        if field_dict is None:
            field_dict = {}
        if exogenous_representation_list is None:
            exogenous_representation_list = []

        self.__source = source
        self.__id = id
        self.__output_directory = output_directory
        self.__field_dict = field_dict
        self.__exogenous_representation_list = exogenous_representation_list
        self.__export_json = export_json

        if not isinstance(self.__exogenous_representation_list, list):
            self.__exogenous_representation_list = [self.__exogenous_representation_list]

        if not isinstance(self.__id, list):
            self.__id = [self.__id]

    @property
    def output_directory(self):
        """
        Getter for the output directory where the produced contents will be stored
        """
        return self.__output_directory

    @property
    def id(self) -> List[str]:
        """
        Getter for the id that represents the ids of the produced contents
        """
        return self.__id

    @property
    def source(self) -> RawInformationSource:
        """
        Getter for the raw information source where the original contents are stored
        """
        return self.__source

    @property
    def exogenous_representation_list(self) -> List[ExogenousConfig]:
        """
        Getter for the exogenous_representation_list
        """
        return self.__exogenous_representation_list

    @property
    def export_json(self) -> bool:
        """
        Getter for the export_json parameter
        """
        return self.__export_json

    def get_configs_list(self, field_name: str) -> List[FieldConfig]:
        """
        Method which returns the list of all `FieldConfig` objects specified for the input `field_name` parameter

        Args:
            field_name: Name of the field for which the list of field configs will be retrieved

        Returns:
            List containing all `FieldConfig` objects specified for the input `field_name`
        """
        return [config for config in self.__field_dict[field_name]]

    def get_field_name_list(self) -> List[str]:
        """
        Method which returns a list containing all the fields of the raw source for which at least one `FieldConfig`
        object has been assigned (i.e. at least one complex representations is specified)

        Returns:
            List of all the fields of the raw source that must be complexly represented
        """
        return list(self.__field_dict.keys())

    def add_single_config(self, field_name: str, field_config: FieldConfig):
        """
        Method which adds a single complex representation for the `field_name` of the raw source

        Examples:

            * Represent field "Plot" of the raw source with a tf-idf technique using sklearn
            >>> import clayrs.content_analyzer as ca
            >>> movies_ca_config.add_single_config("Plot", FieldConfig(ca.SkLearnTfIdf()))

        Args:
            field_name: field name of the raw source which must be complexly represented
            field_config: `FieldConfig` specifying how to represent the field of the raw source
        """
        # If the field_name is not in the field_dict keys it means there is no list to append the FieldConfig to,
        # so a new list is instantiated
        if self.__field_dict.get(field_name) is not None:
            self.__field_dict[field_name].append(field_config)
        else:
            self.__field_dict[field_name] = list()
            self.__field_dict[field_name].append(field_config)

    def add_multiple_config(self, field_name: str, config_list: List[FieldConfig]):
        """
        Method which adds multiple complex representations for the `field_name` of the raw source

        Examples:

            * Represent preprocessed field "Plot" of the raw source with a tf-idf technique using sklearn and a word
            embedding technique using Word2Vec. For the latter, no preprocessing operation will be applied
            >>> import clayrs.content_analyzer as ca
            >>> movies_ca_config.add_multiple_config("Plot",
            >>>                                       [FieldConfig(ca.SkLearnTfIdf(),
            >>>                                                    preprocessing=ca.NLTK(stopwords_removal=True)),
            >>>
            >>>                                        FieldConfig(ca.WordEmbeddingTechnique(ca.GensimWord2Vec()))]

        Args:
            field_name: field name of the raw source which must be complexly represented
            config_list: List of `FieldConfig` objects specifying how to represent the field of the raw source
        """
        # If the field_name is not in the field_dict keys it means there is no list to append the FieldConfig to,
        # so a new list is instantiated
        if self.__field_dict.get(field_name) is not None:
            self.__field_dict[field_name].extend(config_list)
        else:
            self.__field_dict[field_name] = list()
            self.__field_dict[field_name].extend(config_list)

    def add_single_exogenous(self, exogenous_config: ExogenousConfig):
        """
        Method which adds a single exogenous representation which will be used to expand each content

        Examples:

            * Expand each content by using DBPedia as external source
            >>> import clayrs.content_analyzer as ca
            >>> movies_ca_config.add_single_exogenous(
            >>>     ca.ExogenousConfig(
            >>>         ca.DBPediaMappingTechnique('dbo:Film', 'Title', 'EN')
            >>>     )
            >>> )

        Args:
            exogenous_config: `ExogenousConfig` object specifying how to expand each content
        """
        self.__exogenous_representation_list.append(exogenous_config)

    def add_multiple_exogenous(self, config_list: List[ExogenousConfig]):
        """
        Method which adds multiple exogenous representations which will be used to expand each content

        Examples:

            * Expand each content by using DBPedia as external source and local dataset as external source
            >>> import clayrs.content_analyzer as ca
            >>> movies_ca_config.add_single_exogenous(
            >>>     [
            >>>         ca.ExogenousConfig(
            >>>             ca.DBPediaMappingTechnique('dbo:Film', 'Title', 'EN')
            >>>         ),
            >>>
            >>>         ca.ExogenousConfig(
            >>>             ca.PropertiesFromDataset(field_name_list=['director'])
            >>>         ),
            >>>     ]
            >>> )

        Args:
            config_list: List containing `ExogenousConfig` objects specifying how to expand each content
        """
        self.__exogenous_representation_list.extend(config_list)

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        raise NotImplementedError


class UserAnalyzerConfig(ContentAnalyzerConfig):
    """
    Class that represents the configuration for the content analyzer. The configuration specifies how the
    `Content Analyzer` needs to complexly represent contents, i.e. how to preprocess them and how to represent them
    In particular this class refers to *users*.

    Examples:

        >>> import clayrs.content_analyzer as ca
        >>> raw_source = ca.JSONFile(json_path)
        >>> users_config = ca.UserAnalyzerConfig(raw_source, id='user_id', output_directory='users_codified/')
        >>> # add single field config
        >>> users_config.add_single_config('occupation', FieldConfig(content_technique=ca.OriginalData()))
        >>> # add single exogenous technique
        >>> users_config.add_single_exogenous(ca.ExogenousConfig(ca.PropertiesFromDataset(field_name_list=['gender']))

    """
    def __str__(self):
        return str(self.__id)

    def __repr__(self):
        return f'UserAnalyzerConfig(source={self.__source}, ' \
               f'id={self.__id}, output directory={self.__output_directory}, ' \
               f'field_dict= {self.__field_dict}, exogenous representation={self.__exogenous_representation_list} ' \
               f'export_json={self.__export_json})'


class ItemAnalyzerConfig(ContentAnalyzerConfig):
    """
    Class that represents the configuration for the content analyzer. The configuration specifies how the
    `Content Analyzer` needs to complexly represent contents, i.e. how to preprocess them and how to represent them
    In particular this class refers to *items*.

    Examples:

        >>> import clayrs.content_analyzer as ca
        >>> raw_source = ca.JSONFile(json_path)
        >>> movies_config = ca.ItemAnalyzerConfig(raw_source, id='movie_id', output_directory='movies_codified/')
        >>> # add single field config
        >>> movies_config.add_single_config('occupation', FieldConfig(content_technique=ca.OriginalData()))
        >>> # add single exogenous technique
        >>> movies_config.add_single_exogenous(ca.ExogenousConfig(ca.PropertiesFromDataset(field_name_list=['gender']))
    """

    def __str__(self):
        return str(self.__id)

    def __repr__(self):
        return f'ItemAnalyzerConfig(source={self.__source}, ' \
               f'id={self.__id}, output directory={self.__output_directory}, ' \
               f'field_dict= {self.__field_dict}, exogenous representation={self.__exogenous_representation_list} ' \
               f'export_json={self.__export_json})'



from __future__ import annotations
from abc import abstractmethod
from typing import List, Union, TYPE_CHECKING

import numpy as np
from gensim.models import KeyedVectors

if TYPE_CHECKING:
    from clayrs.content_analyzer.information_processor.information_processor import InformationProcessor
    from clayrs.content_analyzer.raw_information_source import RawInformationSource

from clayrs.content_analyzer.embeddings.embedding_source import EmbeddingSource
from clayrs.content_analyzer.utils.check_tokenization import check_tokenized, tokenize_in_sentences, check_not_tokenized
from clayrs.utils.const import logger
from clayrs.utils.context_managers import get_progbar


class EmbeddingLearner(EmbeddingSource):
    """
    Abstract Class for the different kinds of embedding
    The EmbeddingLearner allows to train and store locally embeddings learned from a source on a specific field list.
    So, for example:

        learner = GensimWord2Vec("somedir/word2vec.model")
        learner.fit(JSONFile(file_path), ['Plot', 'Genre'], NLTK())

    The most important method is fit
    It calls the method that creates the model and, if the user wants, it stores the model locally.
    In the example, the model will be fitted on the source located in file_path on the "Plot" and "Genre" fields.
    Preprocessing will also be applied to the text in the source, in this case using NLTK

    Args:
        reference (str): path where the model is stored. If you want to train the model on the spot, pass the path where
            the model will be stored if you save it

            Example:
                I want to load the model locally giving the file_path of the model

                Assume the model file named "example.model" is stored in the dir: "somedir"
                The file_path then will be: "somedir/example.model" or "somedir/example"

                I want to train the model and save it locally

                Assume the dir where the model will be stored is: "somedir"
                The file_path will be as above: "somedir/example.model" or "somedir/example"
                Meaning you have to pass the file name of the model you are going to save

            Alternatively, you can also pass None as reference, in which case, the model won't be saved but it will
            be stored in the Embedding Learner instance (useful if you want to use the trained model only once and then
            discard it)

        auto_save (bool): if True, the model is automatically saved locally after fitting. In particular, you can set
            this to true if you want the model trained during the content analysis process to be saved locally

        extension (str): defines what type of extension the model file will be (for example: .model or .bin). This is
            done so that the user can define a file path with or without the extension (so for example, both
            "somedir/model_file" and "somedir/model_file.model" are acceptable)

        kwargs: arguments that you would pass to any of the models if using them from their original library (for
            example, you can pass "workers=4, min_count=1" to the FastText model)
    """
    def __init__(self, file_path: str, auto_save: bool, extension: str, **kwargs):
        # adds the extension related to the learner if the user didn't pass it
        if file_path is not None and not file_path.endswith(extension):
            file_path += extension

        super().__init__(file_path)

        self._auto_save = auto_save
        self._additional_parameters = kwargs

    @property
    def additional_parameters(self):
        return self._additional_parameters

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    def fit(self, source: RawInformationSource, field_list: List[str],
            preprocessor_list: Union[List[InformationProcessor], InformationProcessor] = None):
        """
        Method that handles the creation and storing of the model
        If the attribute auto_save is True, it automatically stores the model locally after it has been trained

        Args:
            source (RawInformationSource): raw data on which the fitting process will be done
            field_list (List[str]): list of fields to consider from the raw data
            preprocessor_list (Union[List[InformationProcessor], InformationProcessor]): either a list or a single
                information processor that will be used to process the raw data in the fields defined in field list
        """

        if preprocessor_list is None:
            preprocessor_list = []

        if not isinstance(preprocessor_list, list):
            preprocessor_list = [preprocessor_list]

        corpus = self.extract_corpus(source, field_list, preprocessor_list)

        logger.info("Fitting model with extracted corpus...")
        self.fit_model(corpus)

        if self._auto_save and self.reference is not None:
            self.save()

    @abstractmethod
    def fit_model(self, corpus: List):
        """
        This method creates the model, in different ways according to the various implementations.
        The model isn't then returned, but gets stored in the 'model' instance attribute.

        Args:
            corpus (List): data extracted and processed from the raw source which will be used to train the model
        """
        raise NotImplementedError

    def extract_corpus(self, source: RawInformationSource, field_list: List[str],
                       preprocessor_list: List[InformationProcessor]) -> list:
        """
        Extracts the data from the source, from the fields specified in the field_list argument, and processes it
        using the processor_list passed as argument)

        Args:
            source (RawInformationSource): raw data on which the fitting process will be done
            field_list (List[str]): list of fields to consider from the raw data
            preprocessor_list (Union[List[InformationProcessor], InformationProcessor]): either a list or a single
                information processor that will be used to process the raw data in the fields defined in field list

        Returns:
            corpus (list): List of processed data
        """
        corpus = []
        # iter the source
        with get_progbar(list(source)) as pbar:

            for doc in pbar:
                pbar.set_description(f"Preprocessing {', '.join(field_list)} for all contents")
                doc_data = ""
                for field_name in field_list:
                    # apply preprocessing and save the data in the list
                    doc_data += " " + doc[field_name].lower()
                for preprocessor in preprocessor_list:
                    doc_data = preprocessor.process(doc_data)
                corpus.append(self.process_data_granularity(doc_data))
        return corpus

    @abstractmethod
    def process_data_granularity(self, doc_data: str) -> Union[List[str], str]:
        """
        Method that applies modifications to the data in order to fit the granularity of the technique

        Args:
            doc_data (str): data to be modified

        Returns:
            doc_data modified to fit the granularity. For example, if the technique had word granularity and doc_data
            was "this is an example", the output would be ["this", "is", "an", "example"]
        """
        raise NotImplementedError

    @abstractmethod
    def save(self):
        """
        Saves the model in the path stored in the file_path attribute
        """
        raise NotImplementedError


class WordEmbeddingLearner(EmbeddingLearner):
    """
    Extends the EmbeddingLearner to define the embedding learners that work with 'word'
    granularity, meaning that the model expects a list of words as training data
    """

    def __init__(self, reference: str, auto_save: bool, extension: str, **kwargs):
        super().__init__(reference, auto_save, extension, **kwargs)

    def process_data_granularity(self, doc_data: str) -> List[str]:
        return check_tokenized(doc_data)

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def fit_model(self, corpus: List):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, word: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class GensimWordEmbeddingLearner(WordEmbeddingLearner):
    """
    Class that contains the generic behavior of the Gensim models
    """

    def __init__(self, reference: str, auto_save: bool, extension: str, **kwargs):
        super().__init__(reference, auto_save, extension, **kwargs)

    def get_vector_size(self) -> int:
        return self.model.vector_size

    def get_embedding(self, word: str) -> np.ndarray:
        return self.model[word]

    def load_model(self):
        return KeyedVectors.load_word2vec_format(self.reference, binary=True)

    def save(self):
        self.model.save_word2vec_format(self.reference, binary=True)

    @abstractmethod
    def fit_model(self, corpus: List):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class SentenceEmbeddingLearner(EmbeddingLearner):
    """
    Extends the EmbeddingLearner to define the embedding learners that work with
    'sentence' granularity, meaning that the model expects a list of sentences as training data
    """

    def __init__(self, reference: str, auto_save: bool, extension: str, **kwargs):
        super().__init__(reference, auto_save, extension, **kwargs)

    def process_data_granularity(self, doc_data: str) -> List[str]:
        return tokenize_in_sentences(doc_data)

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def fit_model(self, corpus: List):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, sentence: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_token(self, sentence: str):
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class DocumentEmbeddingLearner(EmbeddingLearner):
    """
    Extends the EmbeddingLearner to define the embedding learners that work with
    'document' granularity, meaning that the model expects a list of documents as training data
    """

    def __init__(self, reference: str, auto_save: bool, extension: str, **kwargs):
        super().__init__(reference, auto_save, extension, **kwargs)

    def process_data_granularity(self, doc_data: str) -> str:
        return check_not_tokenized(doc_data)

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def fit_model(self, corpus: List):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_sentence(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_token(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class GensimDocumentEmbeddingLearner(DocumentEmbeddingLearner):
    """
    Class that contains the generic behavior of the Gensim models working at document granularity
    """

    def process_data_granularity(self, doc_data: str) -> str:
        # gensim requires document data to be tokenized in a list
        return check_tokenized(doc_data)

    def get_embedding_sentence(self, document_tokenized: List[str]) -> np.ndarray:
        raise NotImplementedError

    def get_embedding_token(self, document_tokenized: List[str]) -> np.ndarray:
        raise NotImplementedError

    def save(self):
        self.model.save(self.reference)



from abc import abstractmethod
import numpy as np

from clayrs.content_analyzer.embeddings.embedding_source import EmbeddingSource


class EmbeddingLoader(EmbeddingSource):
    """
    Abstract class that generalizes the behavior of the techniques that load an embedding model from a local file or
    download it from the internet. In particular, these techniques can only load the model, they don't provide a method
    to train the model
    """

    def __init__(self, reference: str):
        super().__init__(reference)

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, data) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class WordEmbeddingLoader(EmbeddingLoader):
    """
    Defines the granularity of a loader which will be 'word'
    """

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, word: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class SentenceEmbeddingLoader(EmbeddingLoader):
    """
    Defines the granularity of a loader which will be 'sentence'
    """

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, sentence: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_token(self, sentence: str) -> np.ndarray:
        """
        method that allows me to get embedding for tokens even if we are working with 'sentence' granularity.
        'Attention! not all models can implement this function'

        Args:
            sentence: sentence to analyze

        Returns: matrix in which each row represents the embedding of a token
        """
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class DocumentEmbeddingLoader(EmbeddingLoader):
    """
    Defines the granularity of a loader which will be 'document'
    """

    @abstractmethod
    def load_model(self):
        raise NotImplementedError

    @abstractmethod
    def get_vector_size(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_embedding(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_token(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_embedding_sentence(self, document: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError



import numpy as np
from abc import ABC, abstractmethod


class CombiningTechnique(ABC):
    """
    Class that generalizes the modality in which loaded embeddings will be
    combined to produce a semantic representation.
    """

    @abstractmethod
    def combine(self, embedding_matrix: np.ndarray):
        """
        Combine, in a way specified in the implementations,
        the row of the input matrix

        Args:
            embedding_matrix: matrix whose rows will be combined

        Returns:

        """
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError


class Centroid(CombiningTechnique):
    """
    This class computes the centroid vector of a matrix.
    """
    def combine(self, embedding_matrix: np.ndarray) -> np.ndarray:
        """
        Calculates the centroid of the input matrix

        Args:
            embedding_matrix: np bi-dimensional array where rows are words columns are hidden dimension
                whose centroid will be calculated

        Returns:
            Centroid vector of the input matrix
        """
        return np.nanmean(embedding_matrix, axis=0)

    def __str__(self):
        return "Centroid"

    def __repr__(self):
        return f'Centroid()'


class Sum(CombiningTechnique):
    """
    This class computes the sum vector of a matrix.
    """
    def combine(self, embedding_matrix: np.ndarray) -> np.ndarray:
        """
        Calculates the sum vector of the input matrix

        Args:
            embedding_matrix: np bi-dimensional array where rows are words columns are hidden dimension
                whose sum vector will be calculated

        Returns:
            Sum vector of the input matrix
        """
        return np.sum(embedding_matrix, axis=0)

    def __str__(self):
        return "Sum"

    def __repr__(self):
        return f'Sum()'


class SingleToken(CombiningTechnique):
    """
    Class which takes a specific row as representative of the whole matrix

    Args:
        token_index: index of the row of the matrix to take
    """
    def __init__(self, token_index: int):
        self.token_index = token_index
        super().__init__()

    def combine(self, embedding_matrix: np.ndarray) -> np.ndarray:
        """
        Takes the row with index `token_index` (set in the constructor) from the input `embedding_matrix`

        Args:
            embedding_matrix: np bi-dimensional array where rows are words columns are hidden dimension
                from where the single token will be extracted

        Returns:
            Single row as representative of the whole matrix

        Raises:
            IndexError: Exception raised when `token_index` (set in the constructor) is out of bounds for the input
                matrix
        """
        try:
            sentence_embedding = embedding_matrix[self.token_index]
        except IndexError:
            raise IndexError(f'The embedding matrix has {embedding_matrix.shape[1]} '
                             f'embeddings but you tried to take the {self.token_index+1}th')
        return sentence_embedding

    def __str__(self):
        return "SingleToken"

    def __repr__(self):
        return f"SingleToken(token_index={self.token_index})"
